/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_BLO_DEBUG		FALSE

#include "operator/operator.pmh"

using namespace fe;

void BloatOp::initialize(void)
{
	catalog<String>("Group");
	catalog<String>("Group","label")="Point Group";
	catalog<String>("Group","suggest")="pointGroups";

	catalog<Real>("Envelope")=1.0;
	catalog<bool>("Envelope","paintable")=true;
	catalog<String>("Envelope","hint")="Linearly adjust all effects.";

	catalog<bool>("Weighted")=false;
	catalog<bool>("Weighted","joined")=true;
	catalog<String>("Weighted","hint")=
			"Use locking attribute to resist change.";

	catalog<String>("weightAttr")="bloat";
	catalog<String>("weightAttr","label")="Weight Attribute";
	catalog<String>("weightAttr","enabler")="Weighted";
	catalog<String>("weightAttr","hint")=
			"Name of real attribute describing participation in change"
			" (1.0 is full effect).";

	catalog<Real>("Scaling")=1.0;
	catalog<Real>("Scaling","min")= -100.0;
	catalog<Real>("Scaling","low")= 0.0;
	catalog<Real>("Scaling","high")=2.0;
	catalog<Real>("Scaling","max")=100.0;
	catalog<String>("Scaling","hint")=
			"Scale evenly around center, like a common transform.";

	catalog<Real>("Expansion")=0.0;
	catalog<Real>("Expansion","min")= -100.0;
	catalog<Real>("Expansion","low")= -1.0;
	catalog<Real>("Expansion","high")=1.0;
	catalog<Real>("Expansion","max")=100.0;
	catalog<String>("Expansion","hint")=
			"Scale away from center, with a uniform distance.";

	catalog<Real>("Bloat")=0.0;
	catalog<Real>("Bloat","min")= -100.0;
	catalog<Real>("Bloat","low")= -1.0;
	catalog<Real>("Bloat","high")=1.0;
	catalog<Real>("Bloat","max")=100.0;
	catalog<String>("Bloat","hint")=
			"Scale in direction of normals, with a uniform distance.";

	catalog< sp<Component> >("Input Surface");
}

void BloatOp::handle(Record& a_rSignal)
{
#if FE_BLO_DEBUG
	feLog("BloatOp::handle\n");
#endif

	const Real envelope=catalog<Real>("Envelope");
	const std::vector<Real> envelopePainting=
			catalog< std::vector<Real> >("Envelope","painting");
	const U32 envPaintSize=envelopePainting.size();

	const String weightAttr=catalog<String>("weightAttr");
	const BWORD weighted=(catalog<bool>("Weighted") && !weightAttr.empty());

	const Real bloat=catalog<Real>("Bloat");
	const Real expansion=catalog<Real>("Expansion");
	const Real scaling=catalog<Real>("Scaling");

	const String group=catalog<String>("Group");
	const BWORD grouped=(!group.empty());

	//* output surface is copy of input
	sp<SurfaceAccessibleI> spSurfaceOutput;
	if(!accessOutput(spSurfaceOutput,a_rSignal)) return;

	sp<SurfaceAccessorI> spInputWeight;
	if(weighted)
	{
		access(spInputWeight,"Input Surface",e_point,weightAttr,e_warning);
	}


	sp<SurfaceAccessorI> spOutputPoint;
	if(grouped)
	{
		if(!access(spOutputPoint,spSurfaceOutput,e_pointGroup,group)) return;
	}
	else
	{
		if(!access(spOutputPoint,spSurfaceOutput,e_point,e_position)) return;
	}

	const U32 count=spOutputPoint->count();
	if(!count)
	{
		feLog("BloatOp::handle point count is zero\n");
		catalog<String>("error")="points required";
		return;
	}

#if FE_BLO_DEBUG
	feLog("BloatOp::handle count %d bloat %.6G expansion %.6G scaling %.6G\n",
			count,bloat,expansion,scaling);
#endif

	sp<SurfaceAccessorI> spOutputNormal;
	if(bloat!=0.0)
	{
		access(spOutputNormal,spSurfaceOutput, e_point,e_normal,e_warning);
	}

	BWORD normalWarning=FALSE;

	if(expansion!=0.0 || scaling!=1.0)
	{
		SpatialVector min=spOutputPoint->spatialVector(0);
		SpatialVector max=min;
		for(U32 index=1;index<count;index++)
		{
			const SpatialVector point=spOutputPoint->spatialVector(index);

			for(U32 m=0;m<3;m++)
			{
				const Real value=point[m];
				if(min[m]>value)
				{
					min[m]=value;
				}
				if(max[m]<value)
				{
					max[m]=value;
				}
			}
		}
		const SpatialVector center=0.5*(min+max);

		for(U32 index=0;index<count;index++)
		{
			const SpatialVector point=spOutputPoint->spatialVector(index);
			const I32 pointIndex=grouped? spOutputPoint->integer(index): index;
			const SpatialVector normal=spOutputNormal.isValid()?
					spOutputNormal->spatialVector(pointIndex):
					SpatialVector(0.0,0.0,0.0);
			const Real weight=(spInputWeight.isValid())?
					spInputWeight->real(pointIndex): 1.0;
			const Real fraction=envelope*weight*
					(index<envPaintSize? envelopePainting[pointIndex]: 1.0f);

			const SpatialVector bloated=center+
					((scaling-1.0)*fraction+1.0)*(point-center)+
					fraction*(unitSafe(point-center)*expansion+normal*bloat);

			spOutputPoint->set(index,bloated);

			if(bloat!=0.0 &&
					fabs(normal[0])+fabs(normal[1])+fabs(normal[2])<1e-6)
			{
				normalWarning=TRUE;
			}

#if FE_BLO_DEBUG
			feLog("BloatOp::handle %d/%d %s -> %s\n",index,count,
					c_print(point),c_print(bloated));
#endif
		}
	}
	else
	{
		//* simplified version (normal-based bloat only)
		for(U32 index=0;index<count;index++)
		{
			const SpatialVector point=spOutputPoint->spatialVector(index);
			const I32 pointIndex=grouped? spOutputPoint->integer(index): index;
			const SpatialVector normal=spOutputNormal.isValid()?
					spOutputNormal->spatialVector(pointIndex):
					SpatialVector(0.0,0.0,0.0);
			const Real weight=(spInputWeight.isValid())?
					spInputWeight->real(pointIndex): 1.0;
			const Real fraction=envelope*weight*
					(pointIndex<envPaintSize?
					envelopePainting[pointIndex]: 1.0f);

			if(fabs(normal[0])+fabs(normal[1])+fabs(normal[2])<1e-6)
			{
				normalWarning=TRUE;
			}

			const SpatialVector bloated=point+fraction*normal*bloat;

			spOutputPoint->set(index,bloated);

#if FE_BLO_DEBUG
			feLog("BloatOp::handle simple %d/%d %s -> %s\n",index,count,
					c_print(point),c_print(bloated));
#endif
		}
	}

	if(normalWarning)
	{
		catalog<String>("warning")+="null normals detected;";
	}

#if FE_BLO_DEBUG
	feLog("BloatOp::handle done\n");
#endif
}
